#include <iostream>
#include <string.h>
#include <stdlib.h>
#include <qpdf/QPDF.hh>
#include <qpdf/QPDFPageDocumentHelper.hh>
#include <qpdf/QPDFPageObjectHelper.hh>
#include <qpdf/QUtil.hh>
#include <qpdf/Buffer.hh>
#include <qpdf/QPDFWriter.hh>

// declarations
struct Size {
    double width;
    double height;
};

Size get_size(QPDF &qpdf);

void overlay_page(const std::string &infilename1,
                  const std::string &infilename2,
                  const std::string &outfilename,
                  bool calculate_destination_rect);

// definitions
Size get_size(QPDF &qpdf)
{
    std::vector<QPDFPageObjectHelper> pages =
            QPDFPageDocumentHelper(qpdf).getAllPages();

    // Get size of the first page
    Size size;

    QPDFObjectHandle::Rectangle rect = pages.at(0)
            .getAttribute("/MediaBox", true).getArrayAsRectangle();

    size.height = std::abs(rect.lly - rect.ury);
    size.width = std::abs(rect.llx - rect.urx);

    return size;
}

void overlay_page(const std::string &infilename1,
                  const std::string &infilename2,
                  const std::string &outfilename,
                  bool calculate_destination_rect)
{
    QPDF qpdf_input;
    qpdf_input.processFile(infilename1.c_str());
    qpdf_input.setImmediateCopyFrom(true);

    QPDF qpdf_output;
    qpdf_output.processFile(infilename2.c_str());

    Size input_size = get_size(qpdf_input);
    Size output_size = get_size(qpdf_output);

    QPDFPageObjectHelper input_page =
            QPDFPageDocumentHelper(qpdf_input).getAllPages().back();

    QPDFPageObjectHelper output_page =
            QPDFPageDocumentHelper(qpdf_output).getAllPages().back();

    QPDFObjectHandle page_xobject = qpdf_output.copyForeignObject(
                input_page.getFormXObjectForPage());

    // Find a unique resource name for the new form XObject
    QPDFObjectHandle resources = output_page.getAttribute("/Resources", true);

    int min_suffix = 1000;
    std::string name = resources.getUniqueResourceName("/Fx", min_suffix);

    std::string content;
    if (calculate_destination_rect)
    {
        double x = (output_size.width - input_size.width) / 2;
        double y = output_size.height - input_size.height -
                (output_size.height - input_size.height) / 2;

        content = output_page.placeFormXObject(
                    page_xobject,
                    name,
                    QPDFObjectHandle::Rectangle(
                        x, y,
                        x + input_size.width, y + input_size.height),
                    false);
    }
    else
    {
        content = output_page.placeFormXObject(
                    page_xobject,
                    name,
                    QPDFObjectHandle::Rectangle(
                        0, 0,
                        output_size.width, output_size.height),
                    false);
    }

    if (! content.empty())
    {
        resources.mergeResources(
                    QPDFObjectHandle::parse(
                        "<< /XObject << >> >>"));
        resources.getKey("/XObject")
                .replaceKey(name, page_xobject);
        output_page.addPageContents(
                    QPDFObjectHandle::newStream(
                        &qpdf_output, "q\n"),
                    true);
        output_page.addPageContents(
                    QPDFObjectHandle::newStream(
                        &qpdf_output, "\nQ\n" + content),
                    false);
    }

    QPDFWriter w(qpdf_output, outfilename.c_str());
    w.write();
}

int main(int argc, char* argv[])
{
    overlay_page("a4.pdf", "a3.pdf", "a4_over_a3_without_rect.pdf", false);
    overlay_page("a4.pdf", "a3.pdf", "a4_over_a3_with_rect.pdf", true);
    overlay_page("a4.pdf", "a5.pdf", "a4_over_a5.pdf", false);
    return 0;
}
